﻿namespace StructuralPatterns.Proxy
{
    using System;
    using System.Collections.Concurrent;
    using System.Runtime.Remoting;
    using System.Runtime.Remoting.Channels;
    using System.Runtime.Remoting.Channels.Tcp;
    using System.Threading;

    [Serializable]
    public class EmployeeInfo
    {
        public int Id { get; set; }

        public string FullName { get; set; }

        /* Skipped */
    }

    public interface IEmployeeDataSource
    {
        EmployeeInfo GetEmployeeInfo(int id);

        void SetEmployeeInfo(EmployeeInfo employeeInfo);
    }

    public class EmployeeDataSource : IEmployeeDataSource
    {
        private static ConcurrentDictionary<int, EmployeeInfo> _database
             = new ConcurrentDictionary<int, EmployeeInfo>();

        public EmployeeDataSource()
        {
            Console.WriteLine("EmployeeDataSource ctor ... ");
        }

        public EmployeeInfo GetEmployeeInfo(int id)
        {
            /*  Demo */
            Console.WriteLine("Loading {0} from DB... ", id);
            Thread.Sleep(1000);

            return EmployeeDataSource._database.GetOrAdd(id, this.CreateNewEmployee);
        }

        public void SetEmployeeInfo(EmployeeInfo employeeInfo)
        {
            /* Demo */
            Console.WriteLine("Saving ({0}, {1}) to DB... ",
                employeeInfo.Id, employeeInfo.FullName);

            EmployeeDataSource._database.AddOrUpdate(employeeInfo.Id,
                employeeInfo, (key, value) => employeeInfo);
        }

        private EmployeeInfo CreateNewEmployee(int id)
        {
            return new EmployeeInfo() {
                Id = id,
                FullName = "[NoName]"
            };
        }
    }

    public class EmployeeDataSourceProxy : MarshalByRefObject, IEmployeeDataSource
    {
        #region Singleton implementation

        private static readonly Lazy<EmployeeDataSourceProxy> _instance =
            new Lazy<EmployeeDataSourceProxy>(() => new EmployeeDataSourceProxy());

        public static EmployeeDataSourceProxy Instance
        {
            get { return EmployeeDataSourceProxy._instance.Value; }
        }

        private EmployeeDataSourceProxy()
        {
            Console.WriteLine("EmployeeDataSourceProxy ctor...");
        }

        #endregion

        private readonly IEmployeeDataSource _dataSource = new EmployeeDataSource();

        private static ConcurrentDictionary<int, EmployeeInfo> _cache
            = new ConcurrentDictionary<int, EmployeeInfo>();

        public EmployeeInfo GetEmployeeInfo(int id)
        {
            return EmployeeDataSourceProxy._cache.GetOrAdd(id, this._dataSource.GetEmployeeInfo);
        }

        public void SetEmployeeInfo(EmployeeInfo employieInfo)
        {
            this._dataSource.SetEmployeeInfo(employieInfo);

            EmployeeDataSourceProxy._cache.AddOrUpdate(employieInfo.Id,
                employieInfo, (key, value) => employieInfo);
        }

        public override object InitializeLifetimeService() { return null; }
    }

    public static class DataSourceFactory
    {
        private static bool _isClientNotRegistered = true;
        private static bool _isServerNotRegistered = true;
        private static readonly string _dataSourceUri = "B7167E88-14FC-4023-AF96-CB1E50E7CE5A";
        private static readonly object _threadSafetyObject = new object();

        public static IEmployeeDataSource CreateServerDataSource()
        {
            EmployeeDataSourceProxy dataSource = EmployeeDataSourceProxy.Instance;

            lock (DataSourceFactory._threadSafetyObject) {
                if (DataSourceFactory._isServerNotRegistered) {
                    int tcpPort = 80;

                    TcpServerChannel channel = new TcpServerChannel(tcpPort);
                    ChannelServices.RegisterChannel(channel, true);

                    RemotingConfiguration.RegisterWellKnownServiceType(
                        typeof(EmployeeDataSourceProxy),
                        DataSourceFactory._dataSourceUri, WellKnownObjectMode.Singleton);

                    RemotingServices.Marshal(dataSource,
                        DataSourceFactory._dataSourceUri, typeof(EmployeeDataSourceProxy));

                    DataSourceFactory._isServerNotRegistered = false;
                }
            }

            return dataSource;
        }

        public static IEmployeeDataSource CreateEmployeeDataSource(bool isRemote)
        {
            if (isRemote) {
                return DataSourceFactory.CreateRemoteDataSource();
            }

            return DataSourceFactory.CreateLocalDataSource();
        }

        private static IEmployeeDataSource CreateLocalDataSource()
        {
            return EmployeeDataSourceProxy.Instance;
        }

        private static IEmployeeDataSource CreateRemoteDataSource()
        {
            lock (DataSourceFactory._threadSafetyObject) {
                if (DataSourceFactory._isClientNotRegistered) {
                    TcpClientChannel channel = new TcpClientChannel();
                    ChannelServices.RegisterChannel(channel, true);

                    string server = "localhost";
                    int tcpPort = 80;
                    
                    string serverUrl = string.Format("tcp://{0}:{1}/{2}",
                        server, tcpPort, DataSourceFactory._dataSourceUri);

                    RemotingConfiguration.RegisterWellKnownClientType(
                        typeof(EmployeeDataSourceProxy), serverUrl);

                    DataSourceFactory._isClientNotRegistered = false;
                }
            }

            return EmployeeDataSourceProxy.Instance;
        }
    }
}
